@@ -20,7 +20,7 @@ from __future__ import division, print_function, absolute_import
 import math
 import datetime as dt
 import operator
-from decimal import Decimal, InvalidOperation
+from decimal import Context, Decimal
 from inspect import isclass
 from numbers import Rational
 from fractions import Fraction
@@ -29,13 +29,15 @@ from hypothesis.errors import InvalidArgument, ResolutionFailed
 from hypothesis.control import assume
 from hypothesis._settings import note_deprecation
 from hypothesis.searchstrategy import SearchStrategy
-from hypothesis.internal.compat import ceil, floor, hrange, text_type, \
-    integer_types, get_type_hints, getfullargspec, implements_iterator
+from hypothesis.internal.compat import gcd, ceil, floor, hrange, \
+    text_type, integer_types, get_type_hints, getfullargspec, \
+    implements_iterator
 from hypothesis.internal.floats import is_negative, float_to_int, \
     int_to_float, count_between_floats
+from hypothesis.internal.renaming import renamed_arguments
 from hypothesis.utils.conventions import infer, not_set
 from hypothesis.internal.reflection import proxies, required_args
-from hypothesis.searchstrategy.reprwrapper import ReprWrapperStrategy
+from hypothesis.internal.branchcheck import check_function
 
 __all__ = [
     'nothing',
@@ -44,7 +46,7 @@ __all__ = [
     'choices', 'streaming',
     'booleans', 'integers', 'floats', 'complex_numbers', 'fractions',
     'decimals',
-    'characters', 'text', 'binary', 'uuids',
+    'characters', 'text', 'from_regex', 'binary', 'uuids',
     'tuples', 'lists', 'sets', 'frozensets', 'iterables',
     'dictionaries', 'fixed_dictionaries',
     'sampled_from', 'permutations',
@@ -108,14 +110,24 @@ def cacheable(fn):
     return cached_strategy
 
 
-def defines_strategy(strategy_definition):
-    from hypothesis.searchstrategy.lazy import LazyStrategy
-    _strategies.add(strategy_definition.__name__)
+def base_defines_strategy(force_reusable):
+    def decorator(strategy_definition):
+        from hypothesis.searchstrategy.lazy import LazyStrategy
+        _strategies.add(strategy_definition.__name__)
 
-    @proxies(strategy_definition)
-    def accept(*args, **kwargs):
-        return LazyStrategy(strategy_definition, args, kwargs)
-    return accept
+        @proxies(strategy_definition)
+        def accept(*args, **kwargs):
+            result = LazyStrategy(strategy_definition, args, kwargs)
+            if force_reusable:
+                result.force_has_reusable_values = True
+                assert result.has_reusable_values
+            return result
+        return accept
+    return decorator
+
+
+defines_strategy = base_defines_strategy(False)
+defines_strategy_with_reusable_values = base_defines_strategy(True)
 
 
 class Nothing(SearchStrategy):
@@ -126,6 +138,9 @@ class Nothing(SearchStrategy):
     def do_draw(self, data):
         data.mark_invalid()
 
+    def calc_has_reusable_values(self):
+        return True
+
     def __repr__(self):
         return 'nothing()'
 
@@ -157,10 +172,7 @@ def just(value):
     """
     from hypothesis.searchstrategy.misc import JustStrategy
 
-    def calc_repr():
-        return 'just(%s)' % (repr(value),)
-
-    return ReprWrapperStrategy(JustStrategy(value), calc_repr)
+    return JustStrategy(value)
 
 
 @defines_strategy
@@ -188,7 +200,7 @@ def one_of(*args):
 
 
 @cacheable
-@defines_strategy
+@defines_strategy_with_reusable_values
 def integers(min_value=None, max_value=None):
     """Returns a strategy which generates integers (in Python 2 these may be
     ints or longs).
@@ -247,7 +259,7 @@ def booleans():
 
 
 @cacheable
-@defines_strategy
+@defines_strategy_with_reusable_values
 def floats(
     min_value=None, max_value=None, allow_nan=None, allow_infinity=None
 ):
@@ -274,13 +286,12 @@ def floats(
                     allow_nan
                 ))
 
+    min_value = try_convert(float, min_value, 'min_value')
+    max_value = try_convert(float, max_value, 'max_value')
+
     check_valid_bound(min_value, 'min_value')
     check_valid_bound(max_value, 'max_value')
     check_valid_interval(min_value, max_value, 'min_value', 'max_value')
-    if min_value is not None:
-        min_value = float(min_value)
-    if max_value is not None:
-        max_value = float(max_value)
     if min_value == float(u'-inf'):
         min_value = None
     if max_value == float(u'inf'):
@@ -357,7 +368,7 @@ def floats(
 
 
 @cacheable
-@defines_strategy
+@defines_strategy_with_reusable_values
 def complex_numbers():
     """Returns a strategy that generates complex numbers."""
     from hypothesis.searchstrategy.numbers import ComplexStrategy
@@ -459,7 +470,6 @@ def lists(
 
     if unique_by is not None:
         from hypothesis.searchstrategy.collections import UniqueListStrategy
-        check_strategy(elements)
         min_size = min_size or 0
         max_size = max_size or float(u'inf')
         if average_size is None:
@@ -473,7 +483,6 @@ def lists(
                     _AVERAGE_LIST_LENGTH,
                     min_size * 2
                 )
-        check_valid_sizes(min_size, average_size, max_size)
         result = UniqueListStrategy(
             elements=elements,
             average_size=average_size,
@@ -483,7 +492,6 @@ def lists(
         )
         return result
 
-    check_valid_sizes(min_size, average_size, max_size)
     from hypothesis.searchstrategy.collections import ListStrategy
     if min_size is None:
         min_size = 0
@@ -493,7 +501,6 @@ def lists(
         else:
             average_size = (min_size + max_size) * 0.5
 
-    check_strategy(elements)
     return ListStrategy(
         (elements,), average_length=average_size,
         min_size=min_size, max_size=max_size,
@@ -617,19 +624,19 @@ def streaming(elements):
         Use :func:`data() <hypothesis.strategies.data>` instead.
 
     """
-    check_strategy(elements)
-
     note_deprecation(
         'streaming() has been deprecated. Use the data() strategy instead and '
         'replace stream iteration with data.draw() calls.'
     )
 
+    check_strategy(elements)
+
     from hypothesis.searchstrategy.streams import StreamStrategy
     return StreamStrategy(elements)
 
 
 @cacheable
-@defines_strategy
+@defines_strategy_with_reusable_values
 def characters(whitelist_categories=None, blacklist_categories=None,
                blacklist_characters=None, min_codepoint=None,
                max_codepoint=None, whitelist_characters=None):
@@ -704,7 +711,7 @@ def characters(whitelist_categories=None, blacklist_categories=None,
 
 
 @cacheable
-@defines_strategy
+@defines_strategy_with_reusable_values
 def text(
     alphabet=None,
     min_size=None, average_size=None, max_size=None
@@ -741,6 +748,37 @@ def text(
 
 @cacheable
 @defines_strategy
+def from_regex(regex):
+    """Generates strings that contain a match for the given regex (i.e. ones
+    for which re.search will return a non-None result).
+
+    ``regex`` may be a pattern or :func:`compiled regex <python:re.compile>`.
+    Both byte-strings and unicode strings are supported, and will generate
+    examples of the same type.
+
+    You can use regex flags (such as :const:`python:re.IGNORECASE`,
+    :const:`python:re.DOTALL` or :const:`python:re.UNICODE`) to control
+    generation. Flags can be passed either in compiled regex or inside the
+    pattern with a ``(?iLmsux)`` group.
+
+    Some regular expressions are only partly supported - the underlying
+    strategy checks local matching and relies on filtering to resolve
+    context-dependent expressions.  Using too many of these constructs may
+    cause health-check errors as too many examples are filtered out. This
+    mainly includes (positive or negative) lookahead and lookbehind groups.
+
+    If you want the generated string to match the whole regex you should use
+    boundary markers. So e.g. ``r"\\A.\\Z"`` will return a single character
+    string, while ``"."`` will return any string, and ``r"\\A.$"`` will return
+    a single character optionally followed by a ``"\\n"``.
+
+    """
+    from hypothesis.searchstrategy.regex import regex_strategy
+    return regex_strategy(regex)
+
+
+@cacheable
+@defines_strategy_with_reusable_values
 def binary(
     min_size=None, average_size=None, max_size=None
 ):
@@ -969,36 +1007,87 @@ def from_type(thing):
 
 
 @cacheable
-@defines_strategy
+@defines_strategy_with_reusable_values
 def fractions(min_value=None, max_value=None, max_denominator=None):
     """Returns a strategy which generates Fractions.
 
     If min_value is not None then all generated values are no less than
-    min_value.
+    min_value.  If max_value is not None then all generated values are no
+    greater than max_value.  min_value and max_value may be anything accepted
+    by the `python:`~fractions.Fraction` constructor.
 
-    If max_value is not None then all generated values are no greater than
-    max_value.
-
-    If max_denominator is not None then the absolute value of the denominator
-    of any generated values is no greater than max_denominator. Note that
-    max_denominator must be at least 1.
+    If max_denominator is not None then the denominator of any generated
+    values is no greater than max_denominator. Note that max_denominator must
+    be None or a positive integer.
 
     """
-    check_valid_bound(min_value, 'min_value')
-    check_valid_bound(max_value, 'max_value')
+    min_value = try_convert(Fraction, min_value, 'min_value')
+    max_value = try_convert(Fraction, max_value, 'max_value')
+
     check_valid_interval(min_value, max_value, 'min_value', 'max_value')
-
     check_valid_integer(max_denominator)
-    if max_denominator is not None and max_denominator < 1:
-        raise InvalidArgument(
-            u'Invalid denominator bound %s' % max_denominator
-        )
 
-    denominator_strategy = integers(min_value=1, max_value=max_denominator)
+    if max_denominator is not None:
+        if max_denominator < 1:
+            raise InvalidArgument(
+                'max_denominator=%r must be >= 1' % max_denominator)
+
+        def fraction_bounds(value):
+            """Find the best lower and upper approximation for value."""
+            # Adapted from CPython's Fraction.limit_denominator here:
+            # https://github.com/python/cpython/blob/3.6/Lib/fractions.py#L219
+            if value is None or value.denominator <= max_denominator:
+                return value, value
+            p0, q0, p1, q1 = 0, 1, 1, 0
+            n, d = value.numerator, value.denominator
+            while True:
+                a = n // d
+                q2 = q0 + a * q1
+                if q2 > max_denominator:
+                    break
+                p0, q0, p1, q1 = p1, q1, p0 + a * p1, q2
+                n, d = d, n - a * d
+            k = (max_denominator - q0) // q1
+            low, high = Fraction(p1, q1), Fraction(p0 + k * p1, q0 + k * q1)
+            assert low < value < high
+            return low, high
+
+        # Take the high approximation for min_value and low for max_value
+        bounds = (max_denominator, min_value, max_value)
+        _, min_value = fraction_bounds(min_value)
+        max_value, _ = fraction_bounds(max_value)
+
+        if None not in (min_value, max_value) and min_value > max_value:
+            raise InvalidArgument(
+                'There are no fractions with a denominator <= %r between '
+                'min_value=%r and max_value=%r' % bounds)
+
+    if min_value is not None and min_value == max_value:
+        return just(min_value)
 
     def dm_func(denom):
-        max_num = max_value * denom if max_value is not None else None
-        min_num = min_value * denom if min_value is not None else None
+        """Take denom, construct numerator strategy, and build fraction."""
+        # Four cases of algebra to get integer bounds and scale factor.
+        min_num, max_num = None, None
+        if max_value is None and min_value is None:
+            pass
+        elif min_value is None:
+            max_num = denom * max_value.numerator
+            denom *= max_value.denominator
+        elif max_value is None:
+            min_num = denom * min_value.numerator
+            denom *= min_value.denominator
+        else:
+            low = min_value.numerator * max_value.denominator
+            high = max_value.numerator * min_value.denominator
+            scale = min_value.denominator * max_value.denominator
+            # After calculating our integer bounds and scale factor, we remove
+            # the gcd to avoid drawing more bytes for the example than needed.
+            # Note that `div` can be at most equal to `scale`.
+            div = gcd(scale, gcd(low, high))
+            min_num = denom * low // div
+            max_num = denom * high // div
+            denom *= scale // div
 
         return builds(
             Fraction,
@@ -1006,18 +1095,22 @@ def fractions(min_value=None, max_value=None, max_denominator=None):
             just(denom)
         )
 
-    return denominator_strategy.flatmap(dm_func)
+    if max_denominator is None:
+        return integers(min_value=1).flatmap(dm_func)
+
+    return integers(1, max_denominator).flatmap(dm_func).map(
+        lambda f: f.limit_denominator(max_denominator))
 
 
 @cacheable
-@defines_strategy
+@defines_strategy_with_reusable_values
 def decimals(min_value=None, max_value=None,
              allow_nan=None, allow_infinity=None, places=None):
     """Generates instances of decimals.Decimal, which may be:
 
     - A finite rational number, between ``min_value`` and ``max_value``.
     - Not a Number, if ``allow_nan`` is True.  None means "allow NaN, unless
-      ``min__value`` and ``max_value`` are not None".
+      ``min_value`` and ``max_value`` are not None".
     - Positive or negative infinity, if ``max_value`` and ``min_value``
       respectively are None, and ``allow_infinity`` is not False.  None means
       "allow infinity, unless excluded by the min and max values".
@@ -1037,7 +1130,7 @@ def decimals(min_value=None, max_value=None,
         raise InvalidArgument('places=%r may not be negative' % places)
 
     if min_value is not None:
-        min_value = Decimal(min_value)
+        min_value = try_convert(Decimal, min_value, 'min_value')
         if min_value.is_infinite() and min_value < 0:
             if not (allow_infinity or allow_infinity is None):
                 raise InvalidArgument('allow_infinity=%r, but min_value=%r'
@@ -1047,7 +1140,7 @@ def decimals(min_value=None, max_value=None,
             # This could be positive infinity, quiet NaN, or signalling NaN
             raise InvalidArgument(u'Invalid min_value=%r' % min_value)
     if max_value is not None:
-        max_value = Decimal(max_value)
+        max_value = try_convert(Decimal, max_value, 'max_value')
         if max_value.is_infinite() and max_value > 0:
             if not (allow_infinity or allow_infinity is None):
                 raise InvalidArgument('allow_infinity=%r, but max_value=%r'
@@ -1055,32 +1148,45 @@ def decimals(min_value=None, max_value=None,
             max_value = None
         elif not max_value.is_finite():
             raise InvalidArgument(u'Invalid max_value=%r' % max_value)
-    check_valid_bound(min_value, 'min_value')
-    check_valid_bound(max_value, 'max_value')
     check_valid_interval(min_value, max_value, 'min_value', 'max_value')
 
     if allow_infinity and (None not in (min_value, max_value)):
         raise InvalidArgument('Cannot allow infinity between finite bounds')
-    # Set up a strategy for finite decimals
+    # Set up a strategy for finite decimals.  Note that both floating and
+    # fixed-point decimals require careful handling to remain isolated from
+    # any external precision context - in short, we always work out the
+    # required precision for lossless operation and use context methods.
     if places is not None:
         # Fixed-point decimals are basically integers with a scale factor
+        def ctx(val):
+            """Return a context in which this value is lossless."""
+            precision = ceil(math.log10(abs(val) or 1)) + places + 1
+            return Context(prec=max([precision, 1]))
+
+        def int_to_decimal(val):
+            context = ctx(val)
+            return context.quantize(context.multiply(val, factor), factor)
 
-        def try_quantize(d):
-            try:
-                return d.quantize(factor)
-            except InvalidOperation:  # pragma: no cover
-                return None
         factor = Decimal(10) ** -places
-        max_num = max_value / factor if max_value is not None else None
-        min_num = min_value / factor if min_value is not None else None
-        strat = integers(min_value=min_num, max_value=max_num)\
-            .map(lambda d: try_quantize(d * factor))\
-            .filter(lambda d: d is not None)
+        min_num, max_num = None, None
+        if min_value is not None:
+            min_num = ceil(ctx(min_value).divide(min_value, factor))
+        if max_value is not None:
+            max_num = floor(ctx(max_value).divide(max_value, factor))
+        if None not in (min_num, max_num) and min_num > max_num:
+            raise InvalidArgument(
+                'There are no decimals with %d places between min_value=%r '
+                'and max_value=%r ' % (places, min_value, max_value))
+        strat = integers(min_num, max_num).map(int_to_decimal)
     else:
         # Otherwise, they're like fractions featuring a power of ten
-        strat = fractions(
-            min_value=min_value, max_value=max_value
-        ).map(lambda f: Decimal(f.numerator) / f.denominator)
+        def fraction_to_decimal(val):
+            precision = ceil(math.log10(abs(val.numerator) or 1) +
+                             math.log10(val.denominator)) + 1
+            return Context(prec=precision or 1).divide(
+                Decimal(val.numerator), val.denominator)
+
+        strat = fractions(min_value, max_value).map(fraction_to_decimal)
     # Compose with sampled_from for infinities and NaNs as appropriate
     special = []
     if allow_nan or (allow_nan is None and (None in (min_value, max_value))):
@@ -1139,9 +1245,16 @@ def permutations(values):
     return PermutationStrategy()
 
 
-@defines_strategy
-def datetimes(min_datetime=dt.datetime.min, max_datetime=dt.datetime.max,
-              timezones=none()):
+@defines_strategy_with_reusable_values
+@renamed_arguments(
+    min_datetime='min_value',
+    max_datetime='max_value',
+)
+def datetimes(
+    min_value=dt.datetime.min, max_value=dt.datetime.max,
+    timezones=none(),
+    min_datetime=None, max_datetime=None,
+):
     """A strategy for generating datetimes, which may be timezone-aware.
 
     This strategy works by drawing a naive datetime between ``min_datetime``
@@ -1179,67 +1292,88 @@ def datetimes(min_datetime=dt.datetime.min, max_datetime=dt.datetime.max,
     # more complex API for all users and a useful feature for very few.
     from hypothesis.searchstrategy.datetime import DatetimeStrategy
 
-    check_type(dt.datetime, min_datetime, 'min_datetime')
-    check_type(dt.datetime, max_datetime, 'max_datetime')
-    if min_datetime.tzinfo is not None:
-        raise InvalidArgument('min_datetime=%r must not have tzinfo'
-                              % (min_datetime,))
-    if max_datetime.tzinfo is not None:
-        raise InvalidArgument('max_datetime=%r must not have tzinfo'
-                              % (max_datetime,))
-    check_valid_interval(min_datetime, max_datetime,
-                         'min_datetime', 'max_datetime')
+    check_type(dt.datetime, min_value, 'min_value')
+    check_type(dt.datetime, max_value, 'max_value')
+    if min_value.tzinfo is not None:
+        raise InvalidArgument('min_value=%r must not have tzinfo'
+                              % (min_value,))
+    if max_value.tzinfo is not None:
+        raise InvalidArgument('max_value=%r must not have tzinfo'
+                              % (max_value,))
+    check_valid_interval(min_value, max_value,
+                         'min_value', 'max_value')
     if not isinstance(timezones, SearchStrategy):
         raise InvalidArgument(
             'timezones=%r must be a SearchStrategy that can provide tzinfo '
             'for datetimes (either None or dt.tzinfo objects)' % (timezones,))
-    return DatetimeStrategy(min_datetime, max_datetime, timezones)
+    return DatetimeStrategy(min_value, max_value, timezones)
 
 
-@defines_strategy
-def dates(min_date=dt.date.min, max_date=dt.date.max):
+@defines_strategy_with_reusable_values
+@renamed_arguments(
+    min_date='min_value',
+    max_date='max_value',
+)
+def dates(
+    min_value=dt.date.min, max_value=dt.date.max,
+    min_date=None, max_date=None,
+):
     """A strategy for dates between ``min_date`` and ``max_date``."""
     from hypothesis.searchstrategy.datetime import DateStrategy
 
-    check_type(dt.date, min_date, 'min_date')
-    check_type(dt.date, max_date, 'max_date')
-    check_valid_interval(min_date, max_date, 'min_date', 'max_date')
-    if min_date == max_date:
-        return just(min_date)
-    return DateStrategy(min_date, max_date)
+    check_type(dt.date, min_value, 'min_value')
+    check_type(dt.date, max_value, 'max_value')
+    check_valid_interval(min_value, max_value, 'min_value', 'max_value')
+    if min_value == max_value:
+        return just(min_value)
+    return DateStrategy(min_value, max_value)
 
 
-@defines_strategy
-def times(min_time=dt.time.min, max_time=dt.time.max, timezones=none()):
+@defines_strategy_with_reusable_values
+@renamed_arguments(
+    min_time='min_value',
+    max_time='max_value',
+)
+def times(
+    min_value=dt.time.min, max_value=dt.time.max, timezones=none(),
+    min_time=None, max_time=None,
+):
     """A strategy for times between ``min_time`` and ``max_time``.
 
     The ``timezones`` argument is handled as for :py:func:`datetimes`.
 
     """
-    check_type(dt.time, min_time, 'min_time')
-    check_type(dt.time, max_time, 'max_time')
-    if min_time.tzinfo is not None:
-        raise InvalidArgument('min_time=%r must not have tzinfo' % min_time)
-    if max_time.tzinfo is not None:
-        raise InvalidArgument('max_time=%r must not have tzinfo' % max_time)
-    check_valid_interval(min_time, max_time, 'min_time', 'max_time')
+    check_type(dt.time, min_value, 'min_value')
+    check_type(dt.time, max_value, 'max_value')
+    if min_value.tzinfo is not None:
+        raise InvalidArgument('min_value=%r must not have tzinfo' % min_value)
+    if max_value.tzinfo is not None:
+        raise InvalidArgument('max_value=%r must not have tzinfo' % max_value)
+    check_valid_interval(min_value, max_value, 'min_value', 'max_value')
     day = dt.date(2000, 1, 1)
-    return datetimes(min_datetime=dt.datetime.combine(day, min_time),
-                     max_datetime=dt.datetime.combine(day, max_time),
+    return datetimes(min_value=dt.datetime.combine(day, min_value),
+                     max_value=dt.datetime.combine(day, max_value),
                      timezones=timezones).map(lambda t: t.timetz())
 
 
-@defines_strategy
-def timedeltas(min_delta=dt.timedelta.min, max_delta=dt.timedelta.max):
-    """A strategy for timedeltas between ``min_delta`` and ``max_delta``."""
+@defines_strategy_with_reusable_values
+@renamed_arguments(
+    min_delta='min_value',
+    max_delta='max_value',
+)
+def timedeltas(
+    min_value=dt.timedelta.min, max_value=dt.timedelta.max,
+    min_delta=None, max_delta=None
+):
+    """A strategy for timedeltas between ``min_value`` and ``max_value``."""
     from hypothesis.searchstrategy.datetime import TimedeltaStrategy
 
-    check_type(dt.timedelta, min_delta, 'min_delta')
-    check_type(dt.timedelta, max_delta, 'max_delta')
-    check_valid_interval(min_delta, max_delta, 'min_delta', 'max_delta')
-    if min_delta == max_delta:
-        return just(min_delta)
-    return TimedeltaStrategy(min_delta=min_delta, max_delta=max_delta)
+    check_type(dt.timedelta, min_value, 'min_value')
+    check_type(dt.timedelta, max_value, 'max_value')
+    check_valid_interval(min_value, max_value, 'min_value', 'max_value')
+    if min_value == max_value:
+        return just(min_value)
+    return TimedeltaStrategy(min_value=min_value, max_value=max_value)
 
 
 @cacheable
@@ -1281,8 +1415,6 @@ def composite(f):
                 first_draw = [True]
 
                 def draw(strategy):
-                    if not first_draw[0]:
-                        data.mark_bind()
                     first_draw[0] = False
                     return data.draw(strategy)
 
@@ -1364,6 +1496,7 @@ def choices():
 
 
 @cacheable
+@defines_strategy_with_reusable_values
 def uuids():
     """Returns a strategy that generates UUIDs.
 
@@ -1372,13 +1505,12 @@ def uuids():
 
     """
     from uuid import UUID
-    return ReprWrapperStrategy(
-        shared(randoms(), key='hypothesis.strategies.uuids.generator').map(
-            lambda r: UUID(int=r.getrandbits(128))
-        ), 'uuids()')
+    return shared(randoms(), key='hypothesis.strategies.uuids.generator').map(
+        lambda r: UUID(int=r.getrandbits(128))
+    )
 
 
-@defines_strategy
+@defines_strategy_with_reusable_values
 def runner(default=not_set):
     """A strategy for getting "the current test runner", whatever that may be.
     The exact meaning depends on the entry point, but it will usually be the
@@ -1430,7 +1562,6 @@ def data():
             return 'data(...)'
 
         def draw(self, strategy, label=None):
-            self.data.mark_bind()
             result = self.data.draw(strategy)
             self.count += 1
             if label is not None:
@@ -1501,6 +1632,7 @@ def register_type_strategy(custom_type, strategy):
 # Private API below here
 
 
+@check_function
 def check_type(typ, arg, name=''):
     if name:
         name += '='
@@ -1514,10 +1646,12 @@ def check_type(typ, arg, name=''):
                               % (typ_string, name, arg, type(arg).__name__))
 
 
-def check_strategy(arg):
-    check_type(SearchStrategy, arg)
+@check_function
+def check_strategy(arg, name=''):
+    check_type(SearchStrategy, arg, name)
 
 
+@check_function
 def check_valid_integer(value):
     """Checks that value is either unspecified, or a valid integer.
 
@@ -1529,6 +1663,7 @@ def check_valid_integer(value):
     check_type(integer_types, value)
 
 
+@check_function
 def check_valid_bound(value, name):
     """Checks that value is either unspecified, or a valid interval bound.
 
@@ -1538,9 +1673,31 @@ def check_valid_bound(value, name):
     if value is None or isinstance(value, integer_types + (Rational,)):
         return
     if math.isnan(value):
-        raise InvalidArgument(u'Invalid end point %s %r' % (value, name))
+        raise InvalidArgument(u'Invalid end point %s=%r' % (name, value))
 
 
+@check_function
+def try_convert(typ, value, name):
+    if value is None:
+        return None
+
+    try:
+        return typ(value)
+    except TypeError:
+        raise InvalidArgument(
+            'Cannot convert %s=%r of type %s to type %s' % (
+                name, value, type(value).__name__, typ.__name__
+            )
+        )
+    except ValueError:
+        raise InvalidArgument(
+            'Cannot convert %s=%r to type %s' % (
+                name, value, typ.__name__
+            )
+        )
+
+
+@check_function
 def check_valid_size(value, name):
     """Checks that value is either unspecified, or a valid non-negative size
     expressed as an integer/float.
@@ -1557,6 +1714,7 @@ def check_valid_size(value, name):
         raise InvalidArgument(u'Invalid size %s %r' % (value, name))
 
 
+@check_function
 def check_valid_interval(lower_bound, upper_bound, lower_name, upper_name):
     """Checks that lower_bound and upper_bound are either unspecified, or they
     define a valid interval on the number line.
@@ -1573,6 +1731,7 @@ def check_valid_interval(lower_bound, upper_bound, lower_name, upper_name):
             ))
 
 
+@check_function
 def check_valid_sizes(min_size, average_size, max_size):
     check_valid_size(min_size, 'min_size')
     check_valid_size(max_size, 'max_size')
